using System;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Windows.Forms;

using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace MyDirect3DProject
{
	/// <summary>
	/// This is the main class of my Direct3D application
	/// </summary>
	public class MainClass : Form
	{
		/// <summary>
		/// The rendering device
		/// </summary>
		Device device = null;
		private VertexBuffer vb;
		private IndexBuffer ib;
		
		private long frames;
		private System.DateTime startTime;
		
		public MainClass()
		{
			this.ClientSize = new System.Drawing.Size(640, 480);
			this.Text = "Direct3D Project";
			this.SetStyle(ControlStyles.AllPaintingInWmPaint|ControlStyles.Opaque, true);
		}
		
		public bool InitializeGraphics()
		{
			try {
				startTime = System.DateTime.Now;
				
				// Now let's setup the Direct3D stuff
				PresentParameters presentParams = new PresentParameters();
				presentParams.Windowed   = true;
				presentParams.SwapEffect = SwapEffect.Discard;
				presentParams.EnableAutoDepthStencil = true; // !!!				
				presentParams.AutoDepthStencilFormat = DepthFormat.D16; // !!!
				
				// Create the device
				device = new Device(0, DeviceType.Hardware, this,
				                    CreateFlags.HardwareVertexProcessing, presentParams);
				
				// Setup the event handlers for the device
				device.DeviceLost     += new EventHandler(this.InvalidateDeviceObjects);
				device.DeviceReset    += new EventHandler(this.RestoreDeviceObjects);
				device.Disposing      += new EventHandler(this.DeleteDeviceObjects);
				device.DeviceResizing += new CancelEventHandler(this.EnvironmentResizing);
				
				vb = new VertexBuffer(typeof(CustomVertex.PositionColored),
				                                   8, device, Usage.WriteOnly,
				                                   0, Pool.Managed);
				CustomVertex.PositionColored[] v = (CustomVertex.PositionColored[]) vb.Lock(0, LockFlags.None);
				v[0] = new CustomVertex.PositionColored(-0.5f, +0.5f, +0.5f, Color.Black.ToArgb());
				v[1] = new CustomVertex.PositionColored(-0.5f, -0.5f, +0.5f, Color.Red.ToArgb());
				v[2] = new CustomVertex.PositionColored(+0.5f, +0.5f, +0.5f, Color.Green.ToArgb());
				v[3] = new CustomVertex.PositionColored(+0.5f, -0.5f, +0.5f, Color.Beige.ToArgb());
				v[4] = new CustomVertex.PositionColored(-0.5f, +0.5f, -0.5f, Color.Yellow.ToArgb());
				v[5] = new CustomVertex.PositionColored(+0.5f, +0.5f, -0.5f, Color.Black.ToArgb());
				v[6] = new CustomVertex.PositionColored(-0.5f, -0.5f, -0.5f, Color.Yellow.ToArgb());
				v[7] = new CustomVertex.PositionColored(+0.5f, -0.5f, -0.5f, Color.Black.ToArgb());				
				vb.Unlock();			
				
				ib = new IndexBuffer(typeof(short), 12*3, device, Usage.WriteOnly, Pool.Managed); // !!!
				// Füllen ginge auch mit ib.Lock, wird aber länger: 12*3 Zuweisungen!
				ib.SetData(new short[]{0,1,2, 1,3,2, 4,5,6, 6,5,7, 0,5,4, 0,2,5, 1,6,7, 1,7,3, 0,6,1, 4,6,0, 2,3,7, 5,2,7},
				           0, LockFlags.None);
				
				return true;
			} catch (DirectXException) {
				return false;
			}			
		}
		
		protected virtual void InvalidateDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void RestoreDeviceObjects(object sender, EventArgs e)
		{
		}
		
		protected virtual void DeleteDeviceObjects(object sender, EventArgs e)
		{
			ib.Dispose();
			vb.Dispose();			
		}
		
		protected virtual void EnvironmentResizing(object sender, CancelEventArgs e)
		{
		}
		
		/// <summary>
		/// This method moves the scene
		/// </summary>
		protected virtual void FrameMove()
		{
			// TODO : Frame movement
		}
		
		/// <summary>
		/// This method renders the scene
		/// </summary>
		protected virtual void Render()
		{
			if (device != null) {

				double time = (System.DateTime.Now - startTime).TotalSeconds;
				
				//device.Clear(ClearFlags.Target, Color.Blue, 1.0f, 0);
				device.Clear(ClearFlags.Target|ClearFlags.ZBuffer, Color.Blue, 1.0f, 0); // !!!
				device.BeginScene();

				device.RenderState.Lighting = false;
				device.RenderState.CullMode = Cull.CounterClockwise; // !!! überflüssig, weil Standard
				
				device.VertexFormat = CustomVertex.PositionColored.Format;
				device.SetStreamSource(0, vb, 0);
								
				device.Indices = ib; // !!!
				
				device.Transform.Projection = Matrix.PerspectiveFovLH(0.8f, Width/(float)Height, 0.1f, 50.0f);
				device.Transform.View = Matrix.LookAtLH(new Vector3(0.0f, 0.0f, -3.0f), new Vector3(), new Vector3(0.0f, 1.0f, 0.0f));
				device.Transform.World = Matrix.RotationY(0.8f*(float)time);

				//device.DrawPrimitives(PrimitiveType.TriangleList, 0, 2);	
				device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, 8, 0, 12); // !!!
				
				device.EndScene();
				device.Present();
				
				// unschön: irgendwann nach einer halben Ewigkeit läuft die Variable frames über
				frames ++;
				Text = (frames/time).ToString("###.0") + " fps";
				
				Invalidate();
			}
		}
		
		/// <summary>
		/// Our mainloop
		/// </summary>
		public void Run()
		{
			// While the form is still valid, render and process messages
			while (Created) {
				FrameMove();	
				Render();
				Application.DoEvents();
			}
		}
		
		protected override void OnPaint(PaintEventArgs e)
		{
			this.Render();
		}
		
		protected override void OnKeyPress(KeyPressEventArgs e)
		{
			base.OnKeyPress(e);
			if ((int)e.KeyChar == (int)System.Windows.Forms.Keys.Escape) {
				this.Close(); 
			}
		}
		
		/// <summary>
		/// The main entry point for the application
		/// </summary>
		static void Main()
		{
			using (MainClass mainClass = new MainClass()) {
				if (!mainClass.InitializeGraphics()) {
					MessageBox.Show("Error while initializing Direct3D");
					return;
				}
				mainClass.Show();
				mainClass.Run();
			}
		}
	}
}
